;; Generated package description from guess-language.el  -*- no-byte-compile: t -*-
(define-package "guess-language" "0.0.1.0.20190417.121229" "Robust automatic language detection" '((cl-lib "0.5") (emacs "24") (nadvice "0.1")) :authors '(("Titus von der Malsburg" . "malsburg@posteo.de")) :maintainer '("Titus von der Malsburg" . "malsburg@posteo.de") :url "https://github.com/tmalsburg/guess-language.el")
